/*
 *                            *
 *                           * *
 *                          *   *
 *                     ***************
 *                      * *       * *
 *                       *  MUMPS  *
 *                      * *       * *
 *                     ***************
 *                          *   *
 *                           * *
 *                            *
 *
 *   fmadm.c
 *    FreeM Administration Tool
 *
 *  
 *   Author: John P. Willis <jpw@coherent-logic.com>
 *    Copyright (C) 1998 MUG Deutschland
 *    Copyright (C) 2020 Coherent Logic Development LLC
 *
 *
 *   This file is part of FreeM.
 *
 *   FreeM is free software: you can redistribute it and/or modify
 *   it under the terms of the GNU Affero Public License as published by
 *   the Free Software Foundation, either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   FreeM is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU Affero Public License for more details.
 *
 *   You should have received a copy of the GNU Affero Public License
 *   along with FreeM.  If not, see <https://www.gnu.org/licenses/>.
 *
 **/

#include <sys/types.h>
#include <sys/stat.h>
#include <stddef.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>

#include "fmadm.h"
#include "errmsg.h"
#include "iniconf.h"
#include "init.h"

/* namespace configuration */
char fma_namespace[STRLEN];
char fma_locktab[STRLEN];
char fma_zallotab[STRLEN];
char fma_routine_path[STRLEN];
char fma_global_path[STRLEN];
char fma_journal_path[STRLEN];
char fma_pct_global_path[STRLEN];
char fma_pct_routine_path[STRLEN];
char fma_journal_cut_threshold[STRLEN];

/* miscellaneous global state */
char obj_str[STRLEN];

extern char config_file[4096];

void fm_configure(void);
void fm_write (FILE *file, char *buf);

int main (int argc, char **argv)
{
    char action[STRLEN];    

    short act = -1;
    short obj = -1;

    char **opts;
    int optc = argc - 3;

    int i = 0;
    int j = 1;

    snprintf (config_file, 4096, "%s/freem.conf", SYSCONFDIR);
    
    /* override for fmadm configure */
    if (argc > 1 && strcmp (argv[1], "configure") == 0) fm_configure ();
    
    /* need at least command, action, object, namespace */
    if (argc < 4) return fmadm_usage();

    /* allocate opts array */
    
    /* first dimension */
    if ((opts = (char **) malloc (FMA_MAXARGS * sizeof (char *))) == NULL) {
        fprintf (stderr, "fmadm [FATAL]:  could not acquire memory\n");
        return 1;
    } 

    /* second dimension */
    for (i = 0; i < FMA_MAXARGS; i++) {
        if ((opts[i] = (char *) malloc (STRLEN * sizeof (char *))) == NULL) {
            fprintf (stderr, "fmadm [FATAL]:  could not acquire memory\n");
            return 1;
        } 
    }

    /* copy argv[4] through argv[argc - 1] to opts[1] through opts[argc - 3] */
    
    strncpy (opts[0], argv[0], STRLEN - 1); /* preserve argv[0] */

    for (i = 4; i < argc; i++) {
        if (i > FMA_MAXARGS) return fmadm_usage(); /* bail if we're going to overrun the array */
        strncpy (opts[j++], argv[i], STRLEN - 1);
    }

    strncpy (action, argv[1], STRLEN - 1);
    strncpy (obj_str, argv[2], STRLEN - 1);
    strncpy (fma_namespace, argv[3], STRLEN - 1);

    if (strncmp (action, "list", STRLEN - 1) == 0) act = ACT_LIST;
    else if (strncmp (action, "examine", STRLEN - 1) == 0) act = ACT_EXAMINE;
    else if (strncmp (action, "verify", STRLEN - 1) == 0) act = ACT_VERIFY;
    else if (strncmp (action, "compact", STRLEN - 1) == 0) act = ACT_COMPACT;
    else if (strncmp (action, "repair", STRLEN - 1) == 0) act = ACT_REPAIR;
    else if (strncmp (action, "create", STRLEN - 1) == 0) act = ACT_CREATE;
    else if (strncmp (action, "remove", STRLEN - 1) == 0) act = ACT_REMOVE;
    else if (strncmp (action, "import", STRLEN - 1) == 0) act = ACT_IMPORT;
    else if (strncmp (action, "export", STRLEN - 1) == 0) act = ACT_EXPORT;
    else if (strncmp (action, "backup", STRLEN - 1) == 0) act = ACT_BACKUP;
    else if (strncmp (action, "restore", STRLEN - 1) == 0) act = ACT_RESTORE;
    else if (strncmp (action, "migrate", STRLEN - 1) == 0) act = ACT_MIGRATE;
    else if (strncmp (action, "edit", STRLEN -1) == 0) act = ACT_EDIT;
    else return fmadm_usage();

    if (strncmp (obj_str, "lock", STRLEN - 1) == 0) obj = OBJ_LOCK;
    else if (strncmp (obj_str, "zallocate", STRLEN - 1) == 0) obj = OBJ_ZALLOC;
    else if (strncmp (obj_str, "journal", STRLEN - 1) == 0) obj = OBJ_JOURNAL;
    else if (strncmp (obj_str, "namespace", STRLEN - 1) == 0) obj = OBJ_NAMESPACE;
    else if (strncmp (obj_str, "global", STRLEN - 1) == 0) obj = OBJ_GLOBAL;
    else if (strncmp (obj_str, "routine", STRLEN - 1) == 0) obj = OBJ_ROUTINE;
    else if (strncmp (obj_str, "job", STRLEN - 1) == 0) obj = OBJ_JOB;
    else return fmadm_usage();

    if (get_conf (fma_namespace, "locktable_file", fma_locktab) == FALSE) {
        fprintf (stderr, "fmadm:  cannot determine LOCK table location for namespace %s\n", fma_namespace);
        return 1;
    }

    if (get_conf (fma_namespace, "zalloctab_file", fma_zallotab) == FALSE) {
        fprintf (stderr, "fmadm:  cannot determine ZALLOCATE table location for namespace %s\n", fma_namespace);
        return 1;
    }   

    if (get_conf (fma_namespace, "routines_path", fma_routine_path) == FALSE) {
        fprintf (stderr, "fmadm:  cannot determine routine path for namespace %s\n", fma_namespace);
        return 1;
    }   

    if (get_conf (fma_namespace, "globals_path", fma_global_path) == FALSE) {
        fprintf (stderr, "fmadm:  cannot determine global path for namespace %s\n", fma_namespace);
        return 1;
    }   

    if (get_conf ("SYSTEM", "globals_path", fma_pct_global_path) == FALSE) {
        fprintf (stderr, "fmadm:  cannot determine %% global path for namespace %s\n", "SYSTEM");
        return 1;
    }

    if (get_conf ("SYSTEM", "routines_path", fma_pct_routine_path) == FALSE) {
        fprintf (stderr, "fmadm:  cannot determine %% routine path for namespace %s\n", "SYSTEM");
        return 1;
    }

    if (get_conf ("SYSTEM", "journal_file", fma_journal_path) == FALSE) {
        strcpy (fma_journal_path, "");
    }

    if (get_conf ("SYSTEM", "journal_cut_threshold", fma_journal_cut_threshold) == FALSE) {
        strcpy (fma_journal_cut_threshold, "1073741824");
    }
    
    strcpy (gloplib, fma_pct_global_path);
    stcnv_c2m (gloplib);

    strcpy (glopath, fma_global_path);
    stcnv_c2m (glopath);


    switch (act) {
        case ACT_LIST: return fm_list (obj, optc, opts);
        case ACT_EXAMINE: return fm_examine (obj, optc, opts);
        case ACT_VERIFY: return fm_verify (obj, optc, opts);
        case ACT_COMPACT: return fm_compact (obj, optc, opts);
        case ACT_REPAIR: return fm_repair (obj, optc, opts);
        case ACT_CREATE: return fm_create (obj, optc, opts);
        case ACT_REMOVE: return fm_remove (obj, optc, opts);
        case ACT_IMPORT: return fm_import (obj, optc, opts);
        case ACT_EXPORT: return fm_export (obj, optc, opts);
        case ACT_BACKUP: return fm_backup (obj, optc, opts);
        case ACT_RESTORE: return fm_restore (obj, optc, opts);
        case ACT_MIGRATE: return fm_migrate (obj, optc, opts);
        case ACT_EDIT: return fm_edit (obj, optc, opts);
        default: return fmadm_usage();
    }

    return 0;   /* should never be reached */

} /* main() */

int fmadm_usage (void)
{

    fprintf (stderr, "\nusage:  fmadm <action> <object> <namespace> [OPTIONS]\n\n");

    fprintf (stderr, "        <action> can be one of:\n");
    fprintf (stderr, "            list, examine, verify, compact, repair, create, remove,\n");
    fprintf (stderr, "            import, export, backup, restore, migrate, edit\n\n");

    fprintf (stderr, "        <object> can be one of:\n");
    fprintf (stderr, "            lock, zallocate, journal, namespace, global, routine, job\n\n");

    fprintf (stderr, "    Not all actions are valid for all objects. Please see the FreeM manual\n");
    fprintf (stderr, "    for details on fmadm usage and options.\n\n");

    return 1;

} /* fmadm_usage() */

int fm_list (short object, int optc, char **options)
{

    switch (object) {

        case OBJ_LOCK:
            return fma_locks_list (optc, options);

        case OBJ_ROUTINE:
            return fma_routines_list (optc, options);

        case OBJ_GLOBAL:
            return fma_globals_list (optc, options);

        case OBJ_JOB:
            return fma_jobs_list (optc, options);

        default:
            fprintf (stderr, "fmadm:  'list' is an invalid action for '%s'\n", obj_str);
            return 1;

    }


} /* fm_list() */

int fm_examine (short object, int optc, char **options)
{

    switch (object) {

        case OBJ_ROUTINE:
            return fma_routines_examine (optc, options);

        case OBJ_GLOBAL:
            return fma_globals_examine (optc, options);

        case OBJ_JOB:
            return fma_jobs_examine (optc, options);

        case OBJ_JOURNAL:
            return fma_journals_examine (optc, options);
            
        default:
            fprintf (stderr, "fmadm:  'examine' is an invalid action for '%s'\n", obj_str);
            return 1;

    }

} /* fm_examine() */

int fm_verify (short object, int optc, char **options)
{

    switch (object) {

        case OBJ_GLOBAL:
            return fma_globals_verify (optc, options);
        
        default:
            fprintf (stderr, "fmadm:  'examine' is an invalid action for '%s'\n", obj_str);
            return 1;

    }

} /* fm_verify() */ 

int fm_compact (short object, int optc, char **options)
{

    switch (object) {

        default:
            fprintf (stderr, "fmadm:  'compact' is an invalid action for '%s'\n", obj_str);
            return 1;

    }

} /* fm_compact() */

int fm_repair (short object, int optc, char **options)
{

    switch (object) {

        default:
            fprintf (stderr, "fmadm:  'repair' is an invalid action for '%s'\n", obj_str);
            return 1;

    }

} /* fm_repair() */

int fm_create (short object, int optc, char **options)
{

    switch (object) {

        default:
            fprintf (stderr, "fmadm:  'create' is an invalid action for '%s'\n", obj_str);
            return 1;

    }
} /* fm_create() */

int fm_remove (short object, int optc, char **options)
{

    switch (object) {

        case OBJ_LOCK:
            return fma_locks_remove (optc, options);

        case OBJ_ROUTINE:
            return fma_routines_remove (optc, options);

        case OBJ_GLOBAL:
            return fma_globals_remove (optc, options);

        default:
            fprintf (stderr, "fmadm:  'remove' is an invalid action for '%s'\n", obj_str);
            return 1;

    }

} /* fm_remove() */

int fm_import (short object, int optc, char **options)
{

    switch (object) {

        case OBJ_ROUTINE: 
            return fma_routines_import (optc, options);

        default:
            fprintf (stderr, "fmadm:  'import' is an invalid action for '%s'\n", obj_str);
            return 1;

    }

} /* fm_import() */

int fm_export (short object, int optc, char **options)
{

    switch (object) {

        case OBJ_ROUTINE: 
            return fma_routines_export (optc, options);

        default:
            fprintf (stderr, "fmadm:  'export' is an invalid action for '%s'\n", obj_str);
            return 1;

    }

} /* fm_export() */

int fm_backup (short object, int optc, char **options)
{

    switch (object) {

        case OBJ_ROUTINE: 
            return fma_routines_backup (optc, options);

        default:
            fprintf (stderr, "fmadm:  'backup' is an invalid action for '%s'\n", obj_str);
            return 1;

    }

} /* fm_backup() */

int fm_restore (short object, int optc, char **options)
{

    switch (object) {

        case OBJ_JOURNAL:
            return fma_journals_restore (optc, options);
        
        default:
            fprintf (stderr, "fmadm:  'restore' is an invalid action for '%s'\n", obj_str);
            return 1;

    }

} /* fm_restore() */

int fm_migrate (short object, int optc, char **options)
{

    switch (object) {

        default:
            fprintf (stderr, "fmadm:  'migrate' is an invalid action for '%s'\n", obj_str);
            return 1;

    }

} /* fm_migrate() */

int fm_edit (short object, int optc, char **options)
{

    switch (object) {

        case OBJ_ROUTINE:
            return fma_routines_edit (optc, options);

        default:
            fprintf (stderr, "fmadm:  'edit' is an invalid action for '%s'\n", obj_str);
            return 1;

    }

} /* fm_edit() */

void fm_configure (void)
{

    char sysrtn[4096];
    char sysgbl[4096];
    char usrrtn[4096];
    char usrgbl[4096];

    char locktab[4096];
    char zalloctab[4096];
    char jnlfile[4096];
    char hcpyfile[4096];
    char jnlmode[4];
    char jnlhostid[4096];    
    char jnlcut[4096];
    
    char buf[4096];
    FILE *fp;

    struct stat etcstat;
    int stat_result;
    
    snprintf (sysrtn, 4095, "%s/freem/SYSTEM/routines", LOCALSTATEDIR);
    snprintf (sysgbl, 4095, "%s/freem/SYSTEM/globals", LOCALSTATEDIR);
    snprintf (usrrtn, 4095, "%s/freem/USER/routines", LOCALSTATEDIR);
    snprintf (usrgbl, 4095, "%s/freem/USER/globals", LOCALSTATEDIR);
    snprintf (locktab, 4095, "%s/freem/locktab", LOCALSTATEDIR);
    snprintf (zalloctab, 4095, "%s/freem/zalloctab", LOCALSTATEDIR);
    snprintf (jnlfile, 4095, "/tmp/freem_journal.dat", LOCALSTATEDIR);
    snprintf (hcpyfile, 4095, "/tmp/freem_hardcopy.dat", LOCALSTATEDIR);
    snprintf (jnlmode, 3, "on");
    snprintf (jnlhostid, 4095, "DEFAULT");
    snprintf (jnlcut, 4095, "4294967000");
    
    if (geteuid () != 0) {
        fprintf (stderr, "fmadm:  not superuser\n");
        exit (1);
    }
    
    if (file_exists (config_file)) {
        fprintf (stderr, "fmadm:  '%s' already exists.\n\n", config_file);
        fprintf (stderr, "'fmadm configure' may only be used on a fresh installation of FreeM.\n");
        exit (1);
    }

    /* check for existence of SYSCONFDIR */
    if (stat (SYSCONFDIR, &etcstat) == -1) {
	fprintf (stderr, "fmadm:  creating %s\n", SYSCONFDIR);
	mkdir (SYSCONFDIR, 0755);
    }
   
    fp = fopen (config_file, "a+");

    printf ("\nFreeM Initial Configuration\n");
    printf ("---------------------------\n\n");

    printf ("This utility will create the initial configuration file for ");
    printf ("FreeM in %s.\n\n", config_file);    
    

    printf ("Creating %s... ", config_file); 
    
    snprintf (buf, 4095, "[SYSTEM]");
    fm_write (fp, buf);

    snprintf (buf, 4095, "root=%s/freem/SYSTEM", LOCALSTATEDIR);
    fm_write (fp, buf);

    snprintf (buf, 4095, "routines_path=%s", sysrtn);
    fm_write (fp, buf);

    snprintf (buf, 4095, "globals_path=%s", sysgbl);
    fm_write (fp, buf);

    snprintf (buf, 4095, "locktable_file=%s", locktab);
    fm_write (fp, buf);

    snprintf (buf, 4095, "zalloctab_file=%s", zalloctab);
    fm_write (fp, buf);

    snprintf (buf, 4095, "journal_file=%s", jnlfile);
    fm_write (fp, buf);

    snprintf (buf, 4095, "journal_mode=%s", jnlmode);
    fm_write (fp, buf);

    snprintf (buf, 4095, "journal_host_id=%s", jnlhostid);
    fm_write (fp, buf);

    snprintf (buf, 4095, "journal_cut_threshold=%s", jnlcut);
    fm_write (fp, buf);
    
    snprintf (buf, 4095, "hardcopy_file=%s", hcpyfile);
    fm_write (fp, buf);

    snprintf (buf, 4095, "\n[USER]");
    fm_write (fp, buf);

    snprintf (buf, 4095, "root=%s/freem/USER", LOCALSTATEDIR);
    fm_write (fp, buf);

    snprintf (buf, 4095, "routines_path=%s", usrrtn);
    fm_write (fp, buf);

    snprintf (buf, 4095, "globals_path=%s", usrgbl);
    fm_write (fp, buf);
    
    
    fclose (fp);

    printf ("[OK]\n\n");

/*
    printf ("Setting USER namespace permissions... ");

    snprintf (buf, 4095, "%s/freem/USER/globals", LOCALSTATEDIR);
    chmod (buf, 0777);

    snprintf (buf, 4095, "%s/freem/USER/routines", LOCALSTATEDIR);
    chmod (buf, 0777);

    printf ("[OK]\n");
    printf ("Setting SYSTEM namespace permissions... ");
    
    snprintf (buf, 4095, "%s/freem/SYSTEM/globals", LOCALSTATEDIR);
    chmod (buf, 0755);

    snprintf (buf, 4095, "%s/freem/SYSTEM/routines", LOCALSTATEDIR);
    chmod (buf, 0755);

    printf ("[OK]\n\n\n");
*/
    printf ("FreeM initial configuration is complete.\n\n");

    printf (" USER globals:                   %s\n", usrgbl);
    printf (" USER routines:                  %s\n", usrrtn);
    printf (" SYSTEM globals:                 %s\n", sysgbl);
    printf (" SYSTEM routines:                %s\n", sysrtn);
    printf (" After-image journal:            %s [%s]\n", jnlfile, jnlmode);
    printf (" Journal cut threshold:          %s bytes\n", jnlcut);
    printf (" Distributed journaling host ID: %s\n", jnlhostid);
    printf (" Hardcopy mode output file:      %s\n\n", hcpyfile);
    
    exit (0);
    
} /* fm_configure */

void fm_write (FILE *file, char *buf)
{
    fprintf (file, "%s\n", buf);
}
